@@ -2,13 +2,15 @@
 #![warn(missing_docs)]
 #![warn(rustdoc::missing_crate_level_docs)]
 
-use parking_lot::Mutex;
+use parking_lot::lock_api::RawMutex as RawMutexTrait;
+use parking_lot::RawMutex;
+use std::cell::UnsafeCell;
 use std::mem;
 use std::sync::{Arc, Weak};
 
 /// A RCell holding either an `Arc<T>`, a `Weak<T>` or being `Empty`.
 #[derive(Debug)]
-pub struct RCell<T>(Mutex<ArcState<T>>);
+pub struct RCell<T>(UnsafeCell<ArcState<T>>);
 
 #[derive(Debug)]
 enum ArcState<T> {
@@ -20,31 +22,34 @@ enum ArcState<T> {
 impl<T> RCell<T> {
     /// Creates a new strong (Arc<T>) RCell from the supplied value.
     pub fn new(value: T) -> Self {
-        RCell(Mutex::new(ArcState::Arc(Arc::new(value))))
+        RCell(UnsafeCell::new(ArcState::Arc(Arc::new(value))))
     }
 
     /// Returns 'true' when this RCell contains a strong `Arc<T>`.
     pub fn retained(&self) -> bool {
-        matches!(*self.0.lock(), ArcState::Arc(_))
+        let lock = self.sharded_lock();
+        matches!(self.get_mut(&lock), ArcState::Arc(_))
     }
 
     /// Returns the number of strong references holding an object alive. The returned strong
     /// count is informal only, the result may be appoximate and has race conditions when
     /// other threads modify the refcount concurrently.
     pub fn refcount(&self) -> usize {
-        self.0.lock().refcount()
+        let lock = self.sharded_lock();
+        self.get_mut(&lock).refcount()
     }
 
     /// Tries to upgrade this RCell from Weak<T> to Arc<T>. This means that as long the RCell
     /// is not dropped the associated data won't be either. When successful it returns
     /// Some<Arc<T>> containing the value, otherwise None is returned on failure.
     pub fn retain(&self) -> Option<Arc<T>> {
-        let mut lock = self.0.lock();
-        match &*lock {
+        let lock = self.sharded_lock();
+        let cell = self.get_mut(&lock);
+        match cell {
             ArcState::Arc(arc) => Some(arc.clone()),
             ArcState::Weak(weak) => {
                 if let Some(arc) = weak.upgrade() {
-                    let _ = mem::replace(&mut *lock, ArcState::Arc(arc.clone()));
+                    let _ = mem::replace(cell, ArcState::Arc(arc.clone()));
                     Some(arc)
                 } else {
                     None
@@ -57,16 +62,18 @@ impl<T> RCell<T> {
     /// Downgrades the RCell, any associated value may become dropped when no other references
     /// exist. When no strong reference left remaining this cell becomes Empty.
     pub fn release(&self) {
-        let mut lock = self.0.lock();
-        if let Some(weak) = match &*lock {
-            ArcState::Arc(arc) => Some(Arc::downgrade(&arc)),
+        let lock = self.sharded_lock();
+        let cell = self.get_mut(&lock);
+
+        if let Some(weak) = match cell {
+            ArcState::Arc(arc) => Some(Arc::downgrade(arc)),
             ArcState::Weak(weak) => Some(weak.clone()),
             ArcState::Empty => None,
         } {
             if weak.strong_count() > 0 {
-                let _ = mem::replace(&mut *lock, ArcState::Weak(weak));
+                let _ = mem::replace(cell, ArcState::Weak(weak));
             } else {
-                let _ = mem::replace(&mut *lock, ArcState::Empty);
+                let _ = mem::replace(cell, ArcState::Empty);
             }
         }
     }
@@ -75,18 +82,49 @@ impl<T> RCell<T> {
     /// *any* resource associated with a RCell (potentially member of a struct that lives
     /// longer) in case one knows that it will never be upgraded again.
     pub fn remove(&self) {
-        let _ = mem::replace(&mut *self.0.lock(), ArcState::Empty);
+        let lock = self.sharded_lock();
+        let _ = mem::replace(self.get_mut(&lock), ArcState::Empty);
     }
 
     /// Tries to get an `Arc<T>` from the RCell. This may fail if the RCell was Weak and all
     /// other `Arc's` became dropped.
     pub fn request(&self) -> Option<Arc<T>> {
-        match &*self.0.lock() {
+        let lock = self.sharded_lock();
+        match self.get_mut(&lock) {
             ArcState::Arc(arc) => Some(arc.clone()),
             ArcState::Weak(weak) => weak.upgrade(),
             ArcState::Empty => None,
         }
     }
+
+    // idea borrowed from crossbeam SeqLock
+    fn sharded_mutex(&self) -> &'static RawMutex {
+        const LEN: usize = 97;
+        static LOCKS: [RawMutex; LEN] = [RawMutex::INIT; LEN];
+        &LOCKS[self as *const Self as usize % LEN]
+    }
+
+    // Acquire a global sharded lock with unlock on drop semantics
+    fn sharded_lock(&self) -> RawMutexGuard<T> {
+        self.sharded_mutex().lock();
+        RawMutexGuard(self)
+    }
+
+    // SAFETY: _is_locked_ is intentionally unused, its only there to denote that the lock must been held.
+    #[allow(clippy::mut_from_ref)]
+    fn get_mut(&self, _is_locked_: &RawMutexGuard<T>) -> &mut ArcState<T> {
+        unsafe { &mut *self.0.get() }
+    }
+}
+
+struct RawMutexGuard<'a, T>(&'a RCell<T>);
+impl<T> Drop for RawMutexGuard<'_, T> {
+    fn drop(&mut self) {
+        unsafe {
+            // SAFETY: the guard gurantees that the we have the lock
+            self.0.sharded_mutex().unlock();
+        }
+    }
 }
 
 /// Helper Trait for replacing the content of a RCell with something new.
@@ -98,41 +136,44 @@ pub trait Replace<T> {
 impl<T> Replace<Arc<T>> for RCell<T> {
     /// Replaces the RCell with the supplied `Arc<T>`. The old entry becomes dropped.
     fn replace(&self, arc: Arc<T>) {
-        let _ = mem::replace(&mut *self.0.lock(), ArcState::Arc(arc));
+        let lock = self.sharded_lock();
+        let _ = mem::replace(self.get_mut(&lock), ArcState::Arc(arc));
     }
 }
 
 impl<T> Replace<Weak<T>> for RCell<T> {
     /// Replaces the RCell with the supplied `Weak<T>`. The old entry becomes dropped.
     fn replace(&self, weak: Weak<T>) {
-        let _ = mem::replace(&mut *self.0.lock(), ArcState::Weak(weak));
+        let lock = self.sharded_lock();
+        let _ = mem::replace(self.get_mut(&lock), ArcState::Weak(weak));
     }
 }
 
 impl<T> From<Arc<T>> for RCell<T> {
     /// Creates a new strong RCell with the supplied `Arc<T>`.
     fn from(arc: Arc<T>) -> Self {
-        RCell(Mutex::new(ArcState::Arc(arc)))
+        RCell(UnsafeCell::new(ArcState::Arc(arc)))
     }
 }
 
 impl<T> From<Weak<T>> for RCell<T> {
     /// Creates a new weak RCell with the supplied `Weak<T>`.
     fn from(weak: Weak<T>) -> Self {
-        RCell(Mutex::new(ArcState::Weak(weak)))
+        RCell(UnsafeCell::new(ArcState::Weak(weak)))
     }
 }
 
 impl<T> Default for RCell<T> {
-    /// Creates an RCell that doesnt hold any reference.
+    /// Creates an RCell that doesn't hold any reference.
     fn default() -> Self {
-        RCell(Mutex::new(ArcState::Empty))
+        RCell(UnsafeCell::new(ArcState::Empty))
     }
 }
 
 impl<T> Clone for RCell<T> {
     fn clone(&self) -> Self {
-        RCell(Mutex::new(self.0.lock().clone()))
+        let lock = self.sharded_lock();
+        RCell(UnsafeCell::new(self.get_mut(&lock).clone()))
     }
 }
 
